@@ -1,7 +1,7 @@
 /*
  * weechat-perl.c - perl plugin for WeeChat
  *
- * Copyright (C) 2003-2015 Sébastien Helleu <flashcode@flashtux.org>
+ * Copyright (C) 2003-2021 Sébastien Helleu <flashcode@flashtux.org>
  * Copyright (C) 2005-2008 Emmanuel Bouthenot <kolter@openics.org>
  *
  * This file is part of WeeChat, the extensible chat client.
@@ -17,7 +17,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with WeeChat.  If not, see <http://www.gnu.org/licenses/>.
+ * along with WeeChat.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #undef _
@@ -37,11 +37,30 @@ WEECHAT_PLUGIN_DESCRIPTION(N_("Support of perl scripts"));
 WEECHAT_PLUGIN_AUTHOR("Sébastien Helleu <flashcode@flashtux.org>");
 WEECHAT_PLUGIN_VERSION(WEECHAT_VERSION);
 WEECHAT_PLUGIN_LICENSE(WEECHAT_LICENSE);
-WEECHAT_PLUGIN_PRIORITY(3000);
+WEECHAT_PLUGIN_PRIORITY(4004);
 
 struct t_weechat_plugin *weechat_perl_plugin = NULL;
 
+struct t_plugin_script_data perl_data;
+
+struct t_config_file *perl_config_file = NULL;
+struct t_config_option *perl_config_look_check_license = NULL;
+struct t_config_option *perl_config_look_eval_keep_context = NULL;
+
 int perl_quiet = 0;
+
+struct t_plugin_script *perl_script_eval = NULL;
+int perl_eval_mode = 0;
+int perl_eval_send_input = 0;
+int perl_eval_exec_commands = 0;
+struct t_gui_buffer *perl_eval_buffer = NULL;
+#define PERL_EVAL_SCRIPT                                                \
+    "sub script_perl_eval {\n"                                          \
+    "    eval \"$_[0]\";\n"                                             \
+    "}\n"                                                               \
+    "weechat::register('" WEECHAT_SCRIPT_EVAL_NAME "', '', '1.0', "     \
+    "'" WEECHAT_LICENSE "', 'Evaluation of source code', '', '');\n"
+
 struct t_plugin_script *perl_scripts = NULL;
 struct t_plugin_script *last_perl_script = NULL;
 struct t_plugin_script *perl_current_script = NULL;
@@ -51,6 +70,7 @@ const char *perl_current_script_filename = NULL;
 PerlInterpreter *perl_current_interpreter = NULL;
 #endif /* MULTIPLICITY */
 int perl_quit_or_upgrade = 0;
+char **perl_buffer_output = NULL;
 
 /*
  * string used to execute action "install":
@@ -94,9 +114,21 @@ char *perl_weechat_code =
 #ifndef MULTIPLICITY
     "package %s;"
 #endif /* MULTIPLICITY */
-    "$SIG{__WARN__} = sub { weechat::print('', 'perl\twarning: '.$_[0]) };"
-    "$SIG{__DIE__} = sub { weechat::print('', 'perl\terror: '.$_[0]) };"
-    "do '%s';"
+    "$SIG{__WARN__} = sub { weechat::print('', '%s '.$_[0]); };"
+    "$SIG{__DIE__} = sub { weechat::print('', '%s '.$_[0]); };"
+    "tie(*STDOUT, 'weechat_output');"
+    "tie(*STDERR, 'weechat_output');"
+    "do %s%s%s"
+    "package weechat_output;"
+    "sub TIEHANDLE { bless {}; }"
+    "sub PRINT {"
+    "  weechat::__output__($_[1]);"
+    "}"
+    "sub PRINTF {"
+    "  my $self = shift;"
+    "  my $fmt = shift;"
+    "  weechat::__output__(sprintf($fmt, @_));"
+    "}"
 };
 
 /*
@@ -132,8 +164,7 @@ weechat_perl_hashtable_to_hash (struct t_hashtable *hashtable)
     if (!hash)
         return NULL;
 
-    weechat_hashtable_map_string (hashtable,
-                                  &weechat_perl_hashtable_map_cb,
+    weechat_hashtable_map_string (hashtable, &weechat_perl_hashtable_map_cb,
                                   hash);
 
     return hash;
@@ -186,6 +217,100 @@ weechat_perl_hash_to_hashtable (SV *hash, int size, const char *type_keys,
     return hashtable;
 }
 
+/*
+ * Flushes output.
+ */
+
+void
+weechat_perl_output_flush ()
+{
+    const char *ptr_command;
+    char *temp_buffer, *command;
+    int length;
+
+    if (!*perl_buffer_output[0])
+        return;
+
+    /* if there's no buffer, we catch the output, so there's no flush */
+    if (perl_eval_mode && !perl_eval_buffer)
+        return;
+
+    temp_buffer = strdup (*perl_buffer_output);
+    if (!temp_buffer)
+        return;
+
+    weechat_string_dyn_copy (perl_buffer_output, NULL);
+
+    if (perl_eval_mode)
+    {
+        if (perl_eval_send_input)
+        {
+            if (perl_eval_exec_commands)
+                ptr_command = temp_buffer;
+            else
+                ptr_command = weechat_string_input_for_buffer (temp_buffer);
+            if (ptr_command)
+            {
+                weechat_command (perl_eval_buffer, temp_buffer);
+            }
+            else
+            {
+                length = 1 + strlen (temp_buffer) + 1;
+                command = malloc (length);
+                if (command)
+                {
+                    snprintf (command, length, "%c%s",
+                              temp_buffer[0], temp_buffer);
+                    weechat_command (perl_eval_buffer,
+                                     (command[0]) ? command : " ");
+                    free (command);
+                }
+            }
+        }
+        else
+        {
+            weechat_printf (perl_eval_buffer, "%s", temp_buffer);
+        }
+    }
+    else
+    {
+        /* script (no eval mode) */
+        weechat_printf (
+            NULL,
+            weechat_gettext ("%s: stdout/stderr (%s): %s"),
+            PERL_PLUGIN_NAME,
+            (perl_current_script) ? perl_current_script->name : "?",
+            temp_buffer);
+    }
+
+    free (temp_buffer);
+}
+
+/*
+ * Redirection for stdout and stderr.
+ */
+
+XS (weechat_perl_output)
+{
+    char *msg, *ptr_msg, *ptr_newline;
+    dXSARGS;
+
+    if (items < 1)
+        return;
+
+    msg = SvPV_nolen (ST (0));
+    ptr_msg = msg;
+    while ((ptr_newline = strchr (ptr_msg, '\n')) != NULL)
+    {
+        weechat_string_dyn_concat (perl_buffer_output,
+                                   ptr_msg,
+                                   ptr_newline - ptr_msg);
+        weechat_perl_output_flush ();
+        ptr_msg = ++ptr_newline;
+    }
+    weechat_string_dyn_concat (perl_buffer_output, ptr_msg, -1);
+}
+
 /*
  * Executes a perl function.
  */
@@ -239,14 +364,14 @@ weechat_perl_exec (struct t_plugin_script *script,
             switch (format[i])
             {
                 case 's': /* string */
-                    XPUSHs(sv_2mortal(newSVpv((char *)argv[i], 0)));
+                    XPUSHs (sv_2mortal(newSVpv((char *)argv[i], 0)));
                     break;
                 case 'i': /* integer */
-                    XPUSHs(sv_2mortal(newSViv(*((int *)argv[i]))));
+                    XPUSHs (sv_2mortal(newSViv(*((int *)argv[i]))));
                     break;
                 case 'h': /* hash */
                     hash = weechat_perl_hashtable_to_hash (argv[i]);
-                    XPUSHs(sv_2mortal(newRV_inc((SV *)hash)));
+                    XPUSHs (sv_2mortal(newRV_inc((SV *)hash)));
                     break;
             }
         }
@@ -259,34 +384,47 @@ weechat_perl_exec (struct t_plugin_script *script,
 
     SPAGAIN;
 
+    weechat_perl_output_flush ();
+
     if (SvTRUE (ERRSV))
     {
         weechat_printf (NULL,
                         weechat_gettext ("%s%s: error: %s"),
                         weechat_prefix ("error"), PERL_PLUGIN_NAME,
                         SvPV_nolen (ERRSV));
-        (void) POPs; /* poping the 'undef' */
+        (void) POPs;  /* pop the "undef" */
         mem_err = 0;
     }
     else
     {
         if (count != 1)
         {
-            weechat_printf (NULL,
-                            weechat_gettext ("%s%s: function \"%s\" must "
-                                             "return one valid value (%d)"),
-                            weechat_prefix ("error"), PERL_PLUGIN_NAME,
-                            function, count);
+            if (ret_type != WEECHAT_SCRIPT_EXEC_IGNORE)
+            {
+                weechat_printf (NULL,
+                                weechat_gettext ("%s%s: function \"%s\" must "
+                                                 "return a valid value"),
+                                weechat_prefix ("error"), PERL_PLUGIN_NAME,
+                                function);
+            }
             mem_err = 0;
         }
         else
         {
             if (ret_type == WEECHAT_SCRIPT_EXEC_STRING)
             {
-                ret_s = newSVsv(POPs);
+                ret_s = newSVsv (POPs);
                 ret_value = strdup (SvPV_nolen (ret_s));
                 SvREFCNT_dec (ret_s);
             }
+            else if (ret_type == WEECHAT_SCRIPT_EXEC_POINTER)
+            {
+                ret_s = newSVsv (POPs);
+                ret_value = plugin_script_str2ptr (weechat_perl_plugin,
+                                                   script->name, function,
+                                                   SvPV_nolen (ret_s));
+                SvREFCNT_dec (ret_s);
+            }
             else if (ret_type == WEECHAT_SCRIPT_EXEC_INT)
             {
                 ret_i = malloc (sizeof (*ret_i));
@@ -303,16 +441,27 @@ weechat_perl_exec (struct t_plugin_script *script,
             }
             else
             {
-                weechat_printf (NULL,
-                                weechat_gettext ("%s%s: function \"%s\" is "
-                                                 "internally misused"),
-                                weechat_prefix ("error"), PERL_PLUGIN_NAME,
-                                function);
+                if (ret_type != WEECHAT_SCRIPT_EXEC_IGNORE)
+                {
+                    weechat_printf (
+                        NULL,
+                        weechat_gettext ("%s%s: function \"%s\" must return "
+                                         "a valid value"),
+                        weechat_prefix ("error"), PERL_PLUGIN_NAME,
+                        function);
+                }
                 mem_err = 0;
             }
         }
     }
 
+    if ((ret_type != WEECHAT_SCRIPT_EXEC_IGNORE) && !ret_value)
+    {
+        weechat_printf (NULL,
+                        weechat_gettext ("%s%s: error in function \"%s\""),
+                        weechat_prefix ("error"), PERL_PLUGIN_NAME, function);
+    }
+
     PUTBACK;
     FREETMPS;
     LEAVE;
@@ -339,14 +488,17 @@ weechat_perl_exec (struct t_plugin_script *script,
 /*
  * Loads a perl script.
  *
- * Returns:
- *   1: OK
- *   0: error
+ * If code is NULL, the content of filename is read and executed.
+ * If code is not NULL, it is executed (the file is not read).
+ *
+ * Returns pointer to new registered script, NULL if error.
  */
 
-int
-weechat_perl_load (const char *filename)
+struct t_plugin_script *
+weechat_perl_load (const char *filename, const char *code)
 {
+    char str_warning[512], str_error[512];
+
     struct t_plugin_script temp_script;
     struct stat buf;
     char *perl_code;
@@ -365,12 +517,16 @@ weechat_perl_load (const char *filename)
     temp_script.shutdown_func = NULL;
     temp_script.charset = NULL;
 
-    if (stat (filename, &buf) != 0)
+    if (!code)
     {
-        weechat_printf (NULL,
-                        weechat_gettext ("%s%s: script \"%s\" not found"),
-                        weechat_prefix ("error"), PERL_PLUGIN_NAME, filename);
-        return 0;
+        if (stat (filename, &buf) != 0)
+        {
+            weechat_printf (NULL,
+                            weechat_gettext ("%s%s: script \"%s\" not found"),
+                            weechat_prefix ("error"), PERL_PLUGIN_NAME,
+                            filename);
+            return NULL;
+        }
     }
 
     if ((weechat_perl_plugin->debug >= 2) || !perl_quiet)
@@ -385,7 +541,7 @@ weechat_perl_load (const char *filename)
     perl_registered_script = NULL;
 
 #ifdef MULTIPLICITY
-    perl_current_interpreter = perl_alloc();
+    perl_current_interpreter = perl_alloc ();
 
     if (!perl_current_interpreter)
     {
@@ -393,27 +549,48 @@ weechat_perl_load (const char *filename)
                         weechat_gettext ("%s%s: unable to create new "
                                          "sub-interpreter"),
                         weechat_prefix ("error"), PERL_PLUGIN_NAME);
-        return 0;
+        return NULL;
     }
 
+    snprintf (str_warning, sizeof (str_warning),
+              weechat_gettext ("%s: warning:"),
+              PERL_PLUGIN_NAME);
+    snprintf (str_error, sizeof (str_error),
+              weechat_gettext ("%s: error:"),
+              PERL_PLUGIN_NAME);
+
     PERL_SET_CONTEXT (perl_current_interpreter);
     perl_construct (perl_current_interpreter);
     temp_script.interpreter = (PerlInterpreter *) perl_current_interpreter;
     perl_parse (perl_current_interpreter, weechat_perl_api_init,
                 perl_args_count, perl_args, NULL);
-    length = strlen (perl_weechat_code) - 2 + strlen (filename) + 1;
+    length = strlen (perl_weechat_code) + strlen (str_warning) +
+        strlen (str_error) - 2 + 4 + strlen ((code) ? code : filename) + 4 + 1;
     perl_code = malloc (length);
     if (!perl_code)
-        return 0;
-    snprintf (perl_code, length, perl_weechat_code, filename);
+        return NULL;
+    snprintf (perl_code, length, perl_weechat_code,
+              str_warning,
+              str_error,
+              (code) ? "{\n" : "'",
+              (code) ? code : filename,
+              (code) ? "\n};\n" : "';");
 #else
     snprintf (pkgname, sizeof (pkgname), "%s%d", PKG_NAME_PREFIX, perl_num);
     perl_num++;
-    length = strlen (perl_weechat_code) - 4 + strlen (pkgname) + strlen (filename) + 1;
+    length = strlen (perl_weechat_code) + strlen (str_warning) +
+        strlen (str_error) - 4 + strlen (pkgname) + 4 +
+        strlen ((code) ? code : filename) + 4 + 1;
     perl_code = malloc (length);
     if (!perl_code)
-        return 0;
-    snprintf (perl_code, length, perl_weechat_code, pkgname, filename);
+        return NULL;
+    snprintf (perl_code, length, perl_weechat_code,
+              pkgname,
+              str_warning,
+              str_error,
+              (code) ? "{\n" : "'",
+              (code) ? code : filename,
+              (code) ? "\n};\n" : "';");
 #endif /* MULTIPLICITY */
     eval_pv (perl_code, TRUE);
     free (perl_code);
@@ -428,7 +605,7 @@ weechat_perl_load (const char *filename)
         weechat_printf (NULL,
                         weechat_gettext ("%s%s: error: %s"),
                         weechat_prefix ("error"), PERL_PLUGIN_NAME,
-                        SvPV_nolen(ERRSV));
+                        SvPV_nolen (ERRSV));
 #ifdef MULTIPLICITY
         perl_destruct (perl_current_interpreter);
         perl_free (perl_current_interpreter);
@@ -441,7 +618,7 @@ weechat_perl_load (const char *filename)
             perl_current_script = NULL;
         }
 
-        return 0;
+        return NULL;
     }
 
     if (!perl_registered_script)
@@ -454,7 +631,7 @@ weechat_perl_load (const char *filename)
         perl_destruct (perl_current_interpreter);
         perl_free (perl_current_interpreter);
 #endif /* MULTIPLICITY */
-        return 0;
+        return NULL;
     }
     perl_current_script = perl_registered_script;
 
@@ -476,7 +653,7 @@ weechat_perl_load (const char *filename)
                                      WEECHAT_HOOK_SIGNAL_STRING,
                                      perl_current_script->filename);
 
-    return 1;
+    return perl_current_script;
 }
 
 /*
@@ -489,7 +666,7 @@ weechat_perl_load_cb (void *data, const char *filename)
     /* make C compiler happy */
     (void) data;
 
-    weechat_perl_load (filename);
+    weechat_perl_load (filename, NULL);
 }
 
 /*
@@ -621,7 +798,7 @@ weechat_perl_reload_name (const char *name)
                                 weechat_gettext ("%s: script \"%s\" unloaded"),
                                 PERL_PLUGIN_NAME, name);
             }
-            weechat_perl_load (filename);
+            weechat_perl_load (filename, NULL);
             free (filename);
         }
     }
@@ -633,19 +810,79 @@ weechat_perl_reload_name (const char *name)
     }
 }
 
+/*
+ * Evaluates perl source code.
+ *
+ * Returns:
+ *   1: OK
+ *   0: error
+ */
+
+int
+weechat_perl_eval (struct t_gui_buffer *buffer, int send_to_buffer_as_input,
+                   int exec_commands, const char *code)
+{
+    void *func_argv[1], *result;
+
+    if (!perl_script_eval)
+    {
+        perl_quiet = 1;
+        perl_script_eval = weechat_perl_load (WEECHAT_SCRIPT_EVAL_NAME,
+                                              PERL_EVAL_SCRIPT);
+        perl_quiet = 0;
+        if (!perl_script_eval)
+            return 0;
+    }
+
+    weechat_perl_output_flush ();
+
+    perl_eval_mode = 1;
+    perl_eval_send_input = send_to_buffer_as_input;
+    perl_eval_exec_commands = exec_commands;
+    perl_eval_buffer = buffer;
+
+    func_argv[0] = (char *)code;
+    result = weechat_perl_exec (perl_script_eval,
+                                WEECHAT_SCRIPT_EXEC_IGNORE,
+                                "script_perl_eval",
+                                "s", func_argv);
+    /* result is ignored */
+    if (result)
+        free (result);
+
+    weechat_perl_output_flush ();
+
+    perl_eval_mode = 0;
+    perl_eval_send_input = 0;
+    perl_eval_exec_commands = 0;
+    perl_eval_buffer = NULL;
+
+    if (!weechat_config_boolean (perl_config_look_eval_keep_context))
+    {
+        perl_quiet = 1;
+        weechat_perl_unload (perl_script_eval);
+        perl_quiet = 0;
+        perl_script_eval = NULL;
+    }
+
+    return 1;
+}
+
 /*
  * Callback for command "/perl".
  */
 
 int
-weechat_perl_command_cb (void *data, struct t_gui_buffer *buffer,
+weechat_perl_command_cb (const void *pointer, void *data,
+                         struct t_gui_buffer *buffer,
                          int argc, char **argv, char **argv_eol)
 {
-    char *ptr_name, *path_script;
+    char *ptr_name, *ptr_code, *path_script;
+    int i, send_to_buffer_as_input, exec_commands;
 
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
-    (void) buffer;
 
     if (argc == 1)
     {
@@ -677,6 +914,10 @@ weechat_perl_command_cb (void *data, struct t_gui_buffer *buffer,
         {
             weechat_perl_unload_all ();
         }
+        else if (weechat_strcasecmp (argv[1], "version") == 0)
+        {
+            plugin_script_display_interpreter (weechat_perl_plugin, 0);
+        }
         else
             WEECHAT_COMMAND_ERROR;
     }
@@ -711,7 +952,8 @@ weechat_perl_command_cb (void *data, struct t_gui_buffer *buffer,
                 /* load perl script */
                 path_script = plugin_script_search_path (weechat_perl_plugin,
                                                          ptr_name);
-                weechat_perl_load ((path_script) ? path_script : ptr_name);
+                weechat_perl_load ((path_script) ? path_script : ptr_name,
+                                   NULL);
                 if (path_script)
                     free (path_script);
             }
@@ -727,6 +969,39 @@ weechat_perl_command_cb (void *data, struct t_gui_buffer *buffer,
             }
             perl_quiet = 0;
         }
+        else if (weechat_strcasecmp (argv[1], "eval") == 0)
+        {
+            send_to_buffer_as_input = 0;
+            exec_commands = 0;
+            ptr_code = argv_eol[2];
+            for (i = 2; i < argc; i++)
+            {
+                if (argv[i][0] == '-')
+                {
+                    if (strcmp (argv[i], "-o") == 0)
+                    {
+                        if (i + 1 >= argc)
+                            WEECHAT_COMMAND_ERROR;
+                        send_to_buffer_as_input = 1;
+                        exec_commands = 0;
+                        ptr_code = argv_eol[i + 1];
+                    }
+                    else if (strcmp (argv[i], "-oc") == 0)
+                    {
+                        if (i + 1 >= argc)
+                            WEECHAT_COMMAND_ERROR;
+                        send_to_buffer_as_input = 1;
+                        exec_commands = 1;
+                        ptr_code = argv_eol[i + 1];
+                    }
+                }
+                else
+                    break;
+            }
+            if (!weechat_perl_eval (buffer, send_to_buffer_as_input,
+                                    exec_commands, ptr_code))
+                WEECHAT_COMMAND_ERROR;
+        }
         else
             WEECHAT_COMMAND_ERROR;
     }
@@ -739,11 +1014,13 @@ weechat_perl_command_cb (void *data, struct t_gui_buffer *buffer,
  */
 
 int
-weechat_perl_completion_cb (void *data, const char *completion_item,
+weechat_perl_completion_cb (const void *pointer, void *data,
+                            const char *completion_item,
                             struct t_gui_buffer *buffer,
                             struct t_gui_completion *completion)
 {
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) completion_item;
     (void) buffer;
@@ -758,9 +1035,11 @@ weechat_perl_completion_cb (void *data, const char *completion_item,
  */
 
 struct t_hdata *
-weechat_perl_hdata_cb (void *data, const char *hdata_name)
+weechat_perl_hdata_cb (const void *pointer, void *data,
+                       const char *hdata_name)
 {
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
 
     return plugin_script_hdata_script (weechat_plugin,
@@ -768,15 +1047,40 @@ weechat_perl_hdata_cb (void *data, const char *hdata_name)
                                        hdata_name);
 }
 
+/*
+ * Returns perl info "perl_eval".
+ */
+
+char *
+weechat_perl_info_eval_cb (const void *pointer, void *data,
+                           const char *info_name,
+                           const char *arguments)
+{
+    char *output;
+
+    /* make C compiler happy */
+    (void) pointer;
+    (void) data;
+    (void) info_name;
+
+    weechat_perl_eval (NULL, 0, 0, (arguments) ? arguments : "");
+    output = strdup (*perl_buffer_output);
+    weechat_string_dyn_copy (perl_buffer_output, NULL);
+
+    return output;
+}
+
 /*
  * Returns infolist with perl scripts.
  */
 
 struct t_infolist *
-weechat_perl_infolist_cb (void *data, const char *infolist_name,
-                          void *pointer, const char *arguments)
+weechat_perl_infolist_cb (const void *pointer, void *data,
+                          const char *infolist_name,
+                          void *obj_pointer, const char *arguments)
 {
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
 
     if (!infolist_name || !infolist_name[0])
@@ -785,7 +1089,7 @@ weechat_perl_infolist_cb (void *data, const char *infolist_name,
     if (weechat_strcasecmp (infolist_name, "perl_script") == 0)
     {
         return plugin_script_infolist_list_scripts (weechat_perl_plugin,
-                                                    perl_scripts, pointer,
+                                                    perl_scripts, obj_pointer,
                                                     arguments);
     }
 
@@ -797,10 +1101,12 @@ weechat_perl_infolist_cb (void *data, const char *infolist_name,
  */
 
 int
-weechat_perl_signal_debug_dump_cb (void *data, const char *signal,
+weechat_perl_signal_debug_dump_cb (const void *pointer, void *data,
+                                   const char *signal,
                                    const char *type_data, void *signal_data)
 {
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) signal;
     (void) type_data;
@@ -814,61 +1120,21 @@ weechat_perl_signal_debug_dump_cb (void *data, const char *signal,
     return WEECHAT_RC_OK;
 }
 
-/*
- * Display infos about external libraries used.
- */
-
-int
-weechat_perl_signal_debug_libs_cb (void *data, const char *signal,
-                                   const char *type_data, void *signal_data)
-{
-    /* make C compiler happy */
-    (void) data;
-    (void) signal;
-    (void) type_data;
-    (void) signal_data;
-
-#ifdef PERL_VERSION_STRING
-    weechat_printf (NULL, "  %s: %s", PERL_PLUGIN_NAME, PERL_VERSION_STRING);
-#else
-    weechat_printf (NULL, "  %s: (?)", PERL_PLUGIN_NAME);
-#endif /* PERL_VERSION_STRING */
-
-    return WEECHAT_RC_OK;
-}
-
-/*
- * Callback called when a buffer is closed.
- */
-
-int
-weechat_perl_signal_buffer_closed_cb (void *data, const char *signal,
-                                      const char *type_data, void *signal_data)
-{
-    /* make C compiler happy */
-    (void) data;
-    (void) signal;
-    (void) type_data;
-
-    if (signal_data)
-        plugin_script_remove_buffer_callbacks (perl_scripts, signal_data);
-
-    return WEECHAT_RC_OK;
-}
-
 /*
  * Timer for executing actions.
  */
 
 int
-weechat_perl_timer_action_cb (void *data, int remaining_calls)
+weechat_perl_timer_action_cb (const void *pointer, void *data,
+                              int remaining_calls)
 {
     /* make C compiler happy */
+    (void) data;
     (void) remaining_calls;
 
-    if (data)
+    if (pointer)
     {
-        if (data == &perl_action_install_list)
+        if (pointer == &perl_action_install_list)
         {
             plugin_script_action_install (weechat_perl_plugin,
                                           perl_scripts,
@@ -877,7 +1143,7 @@ weechat_perl_timer_action_cb (void *data, int remaining_calls)
                                           &perl_quiet,
                                           &perl_action_install_list);
         }
-        else if (data == &perl_action_remove_list)
+        else if (pointer == &perl_action_remove_list)
         {
             plugin_script_action_remove (weechat_perl_plugin,
                                          perl_scripts,
@@ -885,7 +1151,7 @@ weechat_perl_timer_action_cb (void *data, int remaining_calls)
                                          &perl_quiet,
                                          &perl_action_remove_list);
         }
-        else if (data == &perl_action_autoload_list)
+        else if (pointer == &perl_action_autoload_list)
         {
             plugin_script_action_autoload (weechat_perl_plugin,
                                            &perl_quiet,
@@ -901,11 +1167,13 @@ weechat_perl_timer_action_cb (void *data, int remaining_calls)
  */
 
 int
-weechat_perl_signal_script_action_cb (void *data, const char *signal,
+weechat_perl_signal_script_action_cb (const void *pointer, void *data,
+                                      const char *signal,
                                       const char *type_data,
                                       void *signal_data)
 {
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
 
     if (strcmp (type_data, WEECHAT_HOOK_SIGNAL_STRING) == 0)
@@ -916,7 +1184,7 @@ weechat_perl_signal_script_action_cb (void *data, const char *signal,
                                       (const char *)signal_data);
             weechat_hook_timer (1, 0, 1,
                                 &weechat_perl_timer_action_cb,
-                                &perl_action_install_list);
+                                &perl_action_install_list, NULL);
         }
         else if (strcmp (signal, "perl_script_remove") == 0)
         {
@@ -924,7 +1192,7 @@ weechat_perl_signal_script_action_cb (void *data, const char *signal,
                                       (const char *)signal_data);
             weechat_hook_timer (1, 0, 1,
                                 &weechat_perl_timer_action_cb,
-                                &perl_action_remove_list);
+                                &perl_action_remove_list, NULL);
         }
         else if (strcmp (signal, "perl_script_autoload") == 0)
         {
@@ -932,7 +1200,7 @@ weechat_perl_signal_script_action_cb (void *data, const char *signal,
                                       (const char *)signal_data);
             weechat_hook_timer (1, 0, 1,
                                 &weechat_perl_timer_action_cb,
-                                &perl_action_autoload_list);
+                                &perl_action_autoload_list, NULL);
         }
     }
 
@@ -944,11 +1212,13 @@ weechat_perl_signal_script_action_cb (void *data, const char *signal,
  */
 
 int
-weechat_perl_signal_quit_upgrade_cb (void *data, const char *signal,
+weechat_perl_signal_quit_upgrade_cb (const void *pointer, void *data,
+                                     const char *signal,
                                      const char *type_data,
                                      void *signal_data)
 {
     /* make C compiler happy */
+    (void) pointer;
     (void) data;
     (void) signal;
     (void) type_data;
@@ -966,7 +1236,6 @@ weechat_perl_signal_quit_upgrade_cb (void *data, const char *signal,
 int
 weechat_plugin_init (struct t_weechat_plugin *plugin, int argc, char *argv[])
 {
-    struct t_plugin_script_init init;
 #ifdef PERL_SYS_INIT3
     int a;
     char **perl_args_local;
@@ -979,6 +1248,22 @@ weechat_plugin_init (struct t_weechat_plugin *plugin, int argc, char *argv[])
 
     weechat_perl_plugin = plugin;
 
+    /* set interpreter name and version */
+    weechat_hashtable_set (plugin->variables, "interpreter_name",
+                           plugin->name);
+#ifdef PERL_VERSION_STRING
+    weechat_hashtable_set (plugin->variables, "interpreter_version",
+                           PERL_VERSION_STRING);
+#else
+    weechat_hashtable_set (plugin->variables, "interpreter_version",
+                           "");
+#endif /* PERL_VERSION_STRING */
+
+    /* init stdout/stderr buffer */
+    perl_buffer_output = weechat_string_dyn_alloc (256);
+    if (!perl_buffer_output)
+        return WEECHAT_RC_ERROR;
+
 #ifndef MULTIPLICITY
     perl_main = perl_alloc ();
 
@@ -996,25 +1281,32 @@ weechat_plugin_init (struct t_weechat_plugin *plugin, int argc, char *argv[])
                 perl_args, NULL);
 #endif /* MULTIPLICITY */
 
-    init.callback_command = &weechat_perl_command_cb;
-    init.callback_completion = &weechat_perl_completion_cb;
-    init.callback_hdata = &weechat_perl_hdata_cb;
-    init.callback_infolist = &weechat_perl_infolist_cb;
-    init.callback_signal_debug_dump = &weechat_perl_signal_debug_dump_cb;
-    init.callback_signal_debug_libs = &weechat_perl_signal_debug_libs_cb;
-    init.callback_signal_buffer_closed = &weechat_perl_signal_buffer_closed_cb;
-    init.callback_signal_script_action = &weechat_perl_signal_script_action_cb;
-    init.callback_load_file = &weechat_perl_load_cb;
+    perl_data.config_file = &perl_config_file;
+    perl_data.config_look_check_license = &perl_config_look_check_license;
+    perl_data.config_look_eval_keep_context = &perl_config_look_eval_keep_context;
+    perl_data.scripts = &perl_scripts;
+    perl_data.last_script = &last_perl_script;
+    perl_data.callback_command = &weechat_perl_command_cb;
+    perl_data.callback_completion = &weechat_perl_completion_cb;
+    perl_data.callback_hdata = &weechat_perl_hdata_cb;
+    perl_data.callback_info_eval = &weechat_perl_info_eval_cb;
+    perl_data.callback_infolist = &weechat_perl_infolist_cb;
+    perl_data.callback_signal_debug_dump = &weechat_perl_signal_debug_dump_cb;
+    perl_data.callback_signal_script_action = &weechat_perl_signal_script_action_cb;
+    perl_data.callback_load_file = &weechat_perl_load_cb;
+    perl_data.unload_all = &weechat_perl_unload_all;
 
     perl_quiet = 1;
-    plugin_script_init (weechat_perl_plugin, argc, argv, &init);
+    plugin_script_init (weechat_perl_plugin, argc, argv, &perl_data);
     perl_quiet = 0;
 
     plugin_script_display_short_list (weechat_perl_plugin,
                                       perl_scripts);
 
-    weechat_hook_signal ("quit", &weechat_perl_signal_quit_upgrade_cb, NULL);
-    weechat_hook_signal ("upgrade", &weechat_perl_signal_quit_upgrade_cb, NULL);
+    weechat_hook_signal ("quit",
+                         &weechat_perl_signal_quit_upgrade_cb, NULL, NULL);
+    weechat_hook_signal ("upgrade",
+                         &weechat_perl_signal_quit_upgrade_cb, NULL, NULL);
 
     /* init OK */
     return WEECHAT_RC_OK;
@@ -1029,7 +1321,12 @@ weechat_plugin_end (struct t_weechat_plugin *plugin)
 {
     /* unload all scripts */
     perl_quiet = 1;
-    plugin_script_end (plugin, &perl_scripts, &weechat_perl_unload_all);
+    if (perl_script_eval)
+    {
+        weechat_perl_unload (perl_script_eval);
+        perl_script_eval = NULL;
+    }
+    plugin_script_end (plugin, &perl_data);
     perl_quiet = 0;
 
 #ifndef MULTIPLICITY
@@ -1058,6 +1355,7 @@ weechat_plugin_end (struct t_weechat_plugin *plugin)
         free (perl_action_remove_list);
     if (perl_action_autoload_list)
         free (perl_action_autoload_list);
+    weechat_string_dyn_free (perl_buffer_output, 1);
 
     return WEECHAT_RC_OK;
 }
